package com.apg.eas.ems;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.TimerTask;

import javax.jms.JMSException;
import javax.jms.TextMessage;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.MBeanException;
import javax.management.MalformedObjectNameException;
import javax.management.ReflectionException;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.springframework.http.converter.StringHttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.web.client.RestTemplate;

import com.apg.eas.global.properties.GlobalProperties;
import com.apg.eas.resttemplate.BWMachineDetails;
import com.apg.eas.resttemplate.FindDependencyReq;
import com.apg.eas.resttemplate.FindDependencyRes;
import com.apg.eas.resttemplate.FindDependencySubReq;
import com.apg.eas.resttemplate.FindDependencySubRes;
import com.apg.eas.transport.jms.JmsSessionFactory;
import com.tibco.tibjms.admin.QueueInfo;
import com.tibco.tibjms.admin.ServerInfo;
import com.tibco.tibjms.admin.TibjmsAdmin;
import com.tibco.tibjms.admin.TibjmsAdminException;

public class EMSMonitoringAgent extends TimerTask {

	boolean isAgentTimerExecuting = false;
	protected String MFTopicName = null;
	protected String timeStamp;
	static final Logger logger = Logger.getLogger(EMSMonitoringAgent.class);
	
	protected String emsServerURL = null;
	protected String emsUserName = null;
	protected String emsPassword = null;
	protected TibjmsAdmin emsAdmin = null;
	protected boolean connected  = false;
	protected GlobalProperties globalProperties;

	@SuppressWarnings("unchecked")
	public EMSMonitoringAgent(String serverURL, String userName, String password, GlobalProperties globalProperties)
	{
		this.emsServerURL = serverURL;
		this.emsUserName = userName;
		this.emsPassword = password;
		this.MFTopicName = globalProperties.getJmsTopicName();
		this.globalProperties = globalProperties;
		
        //Create the admin connection to the TIBCO EMS server
        try 
        {
        	logger.info("Using Credentials: "+getEmsServerURL()+"|"+getEmsUserName()+"|"+getEmsPassword());
        	emsAdmin = new TibjmsAdmin(getEmsServerURL(),getEmsUserName(),getEmsPassword());
        	emsAdmin.setAutoSave(true);
        	connected = true;
        } 
        catch (TibjmsAdminException e) 
        {
			logger.error("Exception while initiating monitoring agent for EMS Server: "+emsServerURL);
			//e.printStackTrace();
			connected = false;
			
			JSONObject objQueueInfo = new JSONObject();
			objQueueInfo.put("EMSSERVERNAME", getEmsServerURL());
			objQueueInfo.put("EMSSERVERURL", getEmsServerURL());
			objQueueInfo.put("STATUS", "ERROR");
			objQueueInfo.put("MESSAGE", e.toString());
			
			try
			{
				publishJMSMessage(objQueueInfo.toString());
			}
			catch (JMSException e1) 
			{
				e1.printStackTrace();
			}
		}
    }

	public String getEmsServerURL() {
		return this.emsServerURL;
	}

	public String getEmsUserName() {
		return this.emsUserName;
	}
	
	public String getEmsPassword() {
		return this.emsPassword;
	}

	public TibjmsAdmin getEmsAdmin() {
		return this.emsAdmin;
	}
	
	@Override
	public void run() 
	{
		try
		{
			if (!isAgentTimerExecuting) 
			{
				logger.info("Timer task executed for EMS Server: " + this.emsServerURL);
				this.isAgentTimerExecuting = true;
				if(connected)
					startMonitoring();
				this.isAgentTimerExecuting = false;
			} 
			else 
			{
				logger.info("Timer invocation not processed as EMS Server Agent: "+ this.emsServerURL + " is executing earlier timer");
			}
		}
		catch (MalformedObjectNameException | AttributeNotFoundException
				| InstanceNotFoundException | IntrospectionException
				| MBeanException | ReflectionException | IOException | JMSException e)
		{
			e.printStackTrace();
		} catch (TibjmsAdminException e) 
		{
			e.printStackTrace();
		}
	}

	public void startMonitoring() throws MalformedObjectNameException,
			AttributeNotFoundException, InstanceNotFoundException,
			IntrospectionException, MBeanException, ReflectionException,
			IOException, JMSException, TibjmsAdminException 
	{
		//String json = null;
		//json = fetchEmsQueueInfo();
		//if (json != null)
			//publishJMSMessage(json);
		
		timeStamp = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ").format(new Date());
										
		ArrayList<String> jsonList = new ArrayList<String>();
		
		fetchEMSQueueInfoAsList(jsonList);
		fetchEMSConnectionDetails(jsonList);
		
		for (Iterator<String> iterator = jsonList.iterator(); iterator.hasNext();)
		{
			String jsonString= (String) iterator.next();
			if (jsonString != null)
				publishJMSMessage(jsonString);
		}
	}

	@SuppressWarnings("unchecked")
	public String fetchEmsQueueInfo() throws TibjmsAdminException
	{
		QueueInfo[] allQueueInfo = emsAdmin.getQueues();
        int counter = 0;
        JSONArray objArray = new JSONArray();
        
        while (counter < allQueueInfo.length)
        {
        	JSONObject objQueueInfo = new JSONObject();
    		objQueueInfo.put("QueueName", allQueueInfo[counter].getName());
    		objQueueInfo.put("MaxMsgCount", allQueueInfo[counter].getMaxMsgs());
    		objQueueInfo.put("PendingMsgCount", allQueueInfo[counter].getPendingMessageCount());
    		objQueueInfo.put("PendingMsgSize", allQueueInfo[counter].getPendingMessageSize());
    		objQueueInfo.put("ReceiverCount", allQueueInfo[counter].getReceiverCount());
    		//objAllQueueInfo[counter] = objQueueInfo;
    		objArray.add(objQueueInfo);
        	counter++;
        }
        
        JSONObject emsQueueInfo = new JSONObject();
        //String timeStamp = new SimpleDateFormat("yyyy.MM.dd.HH.mm.ss").format(new Date());
        //String timeStamp = new SimpleDateFormat("dd/MMM/yyyy hh:mm:ss").format(new Date());
        emsQueueInfo.put("timestamp", timeStamp);
        emsQueueInfo.put("EMSServerURL", emsServerURL);
        emsQueueInfo.put("Queues",objArray);
        emsQueueInfo.put("MsgType", "EMS Queue Info");
		return emsQueueInfo.toJSONString();
	}
	
	@SuppressWarnings("unchecked")
	public void fetchEMSQueueInfoAsList(ArrayList<String> jsonList) throws TibjmsAdminException
	{
		QueueInfo[] allQueueInfo = emsAdmin.getQueues();
        int counter = 0;
        StringBuilder publishers = null;
        StringBuilder subscribers = null;
        String destName = null;
        
        for (int i = 0; i < allQueueInfo.length; i++) 
        {
        	JSONObject objQueueInfo = new JSONObject();
        	publishers = new StringBuilder();
            subscribers = new StringBuilder();
            
        	destName = allQueueInfo[i].getName();
        	objQueueInfo.put("DestinationName", destName);
        	objQueueInfo.put("EMSServerURL", emsServerURL);
        	objQueueInfo.put("MsgType", "EMS Queue Info");
        	counter++;
        	
        	fetchPubSub(publishers, subscribers, destName);
        	
        	objQueueInfo.put("Publishers", publishers.toString());
        	objQueueInfo.put("Subscribers", subscribers.toString());
        	objQueueInfo.put("MaxMsgCount", allQueueInfo[counter-1].getMaxMsgs());
        	objQueueInfo.put("PendingMsgCount", allQueueInfo[counter-1].getPendingMessageCount());
    		objQueueInfo.put("PendingMsgSize", allQueueInfo[counter-1].getPendingMessageSize());
    		objQueueInfo.put("ReceiverCount", allQueueInfo[counter-1].getReceiverCount());
        	objQueueInfo.put("timestamp", timeStamp);
        	
        	jsonList.add(objQueueInfo.toString());
		}
	}

	@SuppressWarnings({"deprecation", "unchecked" })
	public void fetchEMSConnectionDetails(ArrayList<String> jsonList) throws TibjmsAdminException
	{
		JSONObject objQueueInfo = new JSONObject();
		ServerInfo emsServerInfo = emsAdmin.getInfo();
		
		objQueueInfo.put("EMSSERVERNAME", emsServerInfo.getServerName());
		objQueueInfo.put("EMSSERVERURL", emsServerInfo.getURL());
		
		switch (emsServerInfo.getState()) 
		{
			case ServerInfo.SERVER_ACTIVE:
				objQueueInfo.put("STATUS", "Running");
				objQueueInfo.put("MESSAGE", "EMS server is active.");
			break;
			
			case ServerInfo.SERVER_STANDBY:
				objQueueInfo.put("STATUS", "StandBy");
				objQueueInfo.put("MESSAGE", "EMS server on stand by.");
			break;

			default:
				objQueueInfo.put("Status", "Unkown");
				objQueueInfo.put("MESSAGE", "EMS server status unknown.");
			break;
		}
		jsonList.add(objQueueInfo.toString());
	}
	
	public void fetchPubSub(StringBuilder publishers, StringBuilder subscribers,String destName)
	{
		Map<String, String> vars = new HashMap<String, String>();
		RestTemplate rt = new RestTemplate();
		rt.getMessageConverters().add(new MappingJackson2HttpMessageConverter());
		rt.getMessageConverters().add(new StringHttpMessageConverter());
		String uri = new String(globalProperties.getPubSubUrl());

		FindDependencySubReq subReq = new FindDependencySubReq();
		FindDependencyRes res =null;
		subReq.setActType("JMS");
		subReq.setConURL("");
		subReq.setDestination(destName);//"QA.APG.NAM.UPDATEVENDORSERVICE.REQUEST.Q.V10");
		subReq.setDesType("queue");
		subReq.setDomainName("APG_QA_02");

		FindDependencyReq req = new FindDependencyReq();
		req.setCorrelationID("12344");
		req.setOperationName("getJMSInfo");
		req.setReq(subReq);

		try 
		{
			res =  rt.postForObject(uri, req, FindDependencyRes.class, vars);
		}
		catch (Exception e)
		{
			System.out.println("URL to fetch publisher and subscriber is down and not responding.");
			publishers.append("");
			subscribers.append("");
		}
		
		if (res != null) 
		{
			for (Iterator<FindDependencySubRes> iterator = res.getContentobj().iterator(); iterator.hasNext();) 
			{
				FindDependencySubRes subRes = (FindDependencySubRes) iterator.next();
				
				String prefix = "";
				
				int start = subRes.getAppName().indexOf("--");
				int last = subRes.getAppName().indexOf(",");//subRes.getAppName().length();
				String tempName = subRes.getAppName().substring(start+2, last);
				
				ArrayList<BWMachineDetails> bwMachineLists = subRes.getDepenDtls();
				
				for (Iterator<BWMachineDetails> it = bwMachineLists.iterator(); it.hasNext();)
				{
					BWMachineDetails bwMachineDetails = (BWMachineDetails) it.next();
					 
					if (subRes.getDirection().equals("Publisher"))
					{
						publishers.append(prefix);
						prefix = ",";
						publishers.append(tempName + "-" +bwMachineDetails.getInstName());
					} 
					else
					{
						subscribers.append(prefix);
						prefix = ",";
						subscribers.append(tempName + "-" +bwMachineDetails.getInstName());
					}
				}
			} 
		}
	}
	
	public void publishJMSMessage(String jsonString) throws JMSException 
	{
		TopicSession JMSSession = null;
		TopicPublisher JMSPublisher = null;
		
		try 
		{
			logger.info("Publishing EMS Json string: " + jsonString);
			JMSSession = JmsSessionFactory.createJMSSession();
			JMSPublisher = JMSSession.createPublisher(JMSSession.createTopic(MFTopicName));
			TextMessage jmsMessage = JMSSession.createTextMessage();
			jmsMessage.setText(jsonString);
			JMSPublisher.send(jmsMessage);
			JMSPublisher.close();
			JMSSession.close();
		} catch (JMSException e) {
			logger.info("Received exception: "	+ e.getMessage());
		} catch (NamingException e) {
			logger.info("Received exception: " + e.getMessage());
		} catch (InterruptedException e) {
			logger.info("Received exception: " + e.getMessage());
		} finally {
			if (JMSPublisher != null) JMSPublisher.close();
			if (JMSSession != null) JMSSession.close();
		}
	}
}