package com.apg.eas.jmx;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TimerTask;

import javax.jms.JMSException;
import javax.jms.TextMessage;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.ReflectionException;
import javax.management.openmbean.CompositeData;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.json.simple.JSONObject;

import com.apg.eas.common.JMXPort;
import com.apg.eas.common.Utils;
import com.apg.eas.global.properties.GlobalProperties;
import com.apg.eas.transport.jms.JmsSessionFactory;

public class JMXPortMonitoringAgent extends TimerTask {

	protected String tempJmxUrl = "service:jmx:rmi:///jndi/rmi://@hostname:@port/jmxrmi";
	protected JMXServiceURL JMXurl = null;
	protected JMXConnector JMXcon = null;
	protected String AgentType = null;
	protected String AgentName = null;
	protected String AgentJMXURL = null;
	protected HashMap<String, String> EngineEntityAttributes = null;
	protected Set<ObjectName> BWEngineMemEntityNames = null;
	boolean isAgentTimerExecuting = false;
	protected String MFTopicName = null;
	protected GlobalProperties globalProperties;
	protected List<JMXPort> jmxPorts;
	protected String timeStamp = null;
		
	static final Logger logger = Logger.getLogger(JMXPortMonitoringAgent.class);
	
	public JMXPortMonitoringAgent(List<JMXPort> jmxPorts,GlobalProperties globalProperties) 
			throws IOException, InstanceNotFoundException,IntrospectionException, MalformedObjectNameException,
																 AttributeNotFoundException, ReflectionException, MBeanException,JMSException 
	{
		this.jmxPorts = jmxPorts;
		this.globalProperties = globalProperties;
		this.MFTopicName = globalProperties.getJmxTopicName();
	}

	@Override
	public void run() 
	{
		try 
		{
			if (!isAgentTimerExecuting)
			{
				timeStamp = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSZ").format(new Date());
				logger.info("############################ EXECUTING TIMER TASK ##############################################"+timeStamp);
				this.isAgentTimerExecuting = true;
				startMonitoring();
				this.isAgentTimerExecuting = false;
				logger.info("############################ TIMER TASK EXECUTED ################################################"+timeStamp);
			}
			else
			{
				logger.info("Timer invocation not processed as Agent: " + this.AgentName + " is executing earlier timer");
			}
		} catch (MalformedObjectNameException | AttributeNotFoundException | InstanceNotFoundException | IntrospectionException
				| MBeanException | ReflectionException | IOException | JMSException e) 
		{
			e.printStackTrace();
		}
	}

	@SuppressWarnings("unchecked")
	public void startMonitoring() throws MalformedObjectNameException, AttributeNotFoundException, InstanceNotFoundException,IntrospectionException, MBeanException, ReflectionException,IOException, JMSException 
	{
		List<String> jsons = null;
		String jmxUrl = null;
		boolean connected = false;
		
		for (Iterator<JMXPort> iterator = jmxPorts.iterator(); iterator.hasNext();) 
		{
			jsons = null;
			jsons = new ArrayList<String>();
			JMXurl = null;
			JMXcon = null;
			JMXPort jmxPort = (JMXPort) iterator.next();
			
			String instanceDetails = jmxPort.getTrafile();
			int indexOfLastSlash = instanceDetails.lastIndexOf("/");
			int indexOfTra = instanceDetails.indexOf(".tra");
			String instanceName = instanceDetails.substring(indexOfLastSlash+1,indexOfTra);
			
			jmxUrl = this.tempJmxUrl;
			jmxUrl = jmxUrl.replace("@hostname", jmxPort.getServer());
			jmxUrl = jmxUrl.replace("@port", jmxPort.getJmxport());
			
			try 
			{
				JMXurl = new JMXServiceURL(jmxUrl);
				JMXcon = JMXConnectorFactory.connect(JMXurl, null);
				connected=true;
			}
			catch(Exception e)
			{
				connected=false;
				JSONObject obj = new JSONObject();
				obj.put("timestamp", timeStamp);
				obj.put("ApplicationName", instanceName);
				obj.put("JMXURL", JMXurl);
				obj.put("MsgType", "BWEngineHealth");
				obj.put("Status", "Stopped");
				obj.put("Message", e.toString());
				obj.put("HeapMemMax", "");
				obj.put("HeapMemUsed", "");
				obj.put("Threads", "");
				obj.put("Uptime", "");
				obj.put("Version", "");
				obj.put("ActiveProcessCount", "");
				obj.put("ProcessCount","");
		    	jsons.add(obj.toString());
			}
			
			if(connected)
			{
				fetchBWEngineHealthJson(jsons,jmxPort.getApplication(),jmxUrl,instanceName);
			}
			
			if (jsons != null && jsons.size() > 0)
				publishJMSMessage(jsons);
			
			logger.info("------------------------------------------POLLING HEALTH STATUS FOR APPLICATION : "+jmxPort.getApplication());
		}
	}
	
	public List<String> fetchBWEngineHealthJson(List<String> jsons,String agentName,String agentJMXUrl,String instanceName) throws InstanceNotFoundException, ReflectionException, IOException, MalformedObjectNameException, IntrospectionException
	{
		HashMap<String, String> bwEngineMemAttributesMap = new HashMap<String, String>();
		String queryString = Utils.getBWEngineMemQueryString();
		ObjectName queryon = new ObjectName(queryString);
		JMXcon.connect();
		
		MBeanInfo info = JMXcon.getMBeanServerConnection().getMBeanInfo(queryon);
		MBeanAttributeInfo[] attrs = info.getAttributes();
		HashMap<String, String> temp = new HashMap<String, String>();
		
		for (int idx = 0; idx < attrs.length; idx++)
		{
			MBeanAttributeInfo attr = attrs[idx];
			temp.put(attr.getName(), null);
		}
		
		String[] attributeNames = (String[]) temp.keySet().toArray(new String[temp.size()]);

		List<Object> list = JMXcon.getMBeanServerConnection().getAttributes(queryon, attributeNames);
		Attribute attr;
		Iterator<Object> iterator = list.iterator();
		
		while (iterator.hasNext()) 
		{
			attr = (Attribute) iterator.next();
			if (attr.getValue() instanceof CompositeData) 
			{
				CompositeData data = (CompositeData)attr.getValue();
				for (String key : data.getCompositeType().keySet()) 
				{
					bwEngineMemAttributesMap.put(attr.getName()+"-"+key, data.get(key).toString());
		        }
			} 
			else
			{
				bwEngineMemAttributesMap.put(attr.getName(), attr.getValue().toString());
			}
		}
		
		String objectName = "com.tibco.bw:key=engine,name=\"" + instanceName + "\"";
		//"com.tibco.bw:key=engine,name=\"APG_Dispatch_BiztalkAcks-APG_Dispatch_BiztalkAcksProcessArchive-LB1\"";
		
		String opGetExecInfo = "GetExecInfo";
		String opGetActiveProcessCount = "GetActiveProcessCount";
		String opGetProcessCount = "GetProcessCount";
		Object opParams[] = null;
		String opSig[] = null;
		CompositeData cd = null;
		try
		{
			Object mbean = invokeMbeanOperation(objectName, opGetExecInfo, opParams, opSig);
			cd = (CompositeData) mbean;
			String status = (String) cd.get("Status");
			
			if(status.equalsIgnoreCase("ACTIVE"))
			{
				bwEngineMemAttributesMap.put("Status", "Running");
			}
			else
			{
				bwEngineMemAttributesMap.put("Status", "Stopped");
			}
			
			bwEngineMemAttributesMap.put("Threads", cd.get("Threads").toString());
			bwEngineMemAttributesMap.put("Uptime", cd.get("Uptime").toString());
			bwEngineMemAttributesMap.put("Version", cd.get("Version").toString());
			
			mbean = invokeMbeanOperation(objectName, opGetActiveProcessCount, opParams, opSig);
			Integer activeProcessCount = (Integer) mbean;
			bwEngineMemAttributesMap.put("ActiveProcessCount", activeProcessCount.toString());
			
			mbean = invokeMbeanOperation(objectName, opGetProcessCount, opParams, opSig);
			Integer processCount = (Integer) mbean;
			bwEngineMemAttributesMap.put("ProcessCount", processCount.toString());
		}
		catch(MalformedObjectNameException  | InstanceNotFoundException | MBeanException | ReflectionException | IOException e) 
		{
			e.printStackTrace();
			
			bwEngineMemAttributesMap.put("Status", "Error");
			bwEngineMemAttributesMap.put("Threads", "");
			bwEngineMemAttributesMap.put("Uptime", "");
			bwEngineMemAttributesMap.put("Version", "");
			bwEngineMemAttributesMap.put("ActiveProcessCount", "");
			bwEngineMemAttributesMap.put("ProcessCount", "");
			bwEngineMemAttributesMap.put("Message", e.getMessage());
		}
		String jsonString = buildFinalJSON(bwEngineMemAttributesMap,instanceName,agentJMXUrl);
		jsons.add(jsonString);
		return jsons;
	}
	
	
	
	public void resetEngineEntityAttributes() {
		Set<String> keySet = this.EngineEntityAttributes.keySet();
		Iterator<String> it = keySet.iterator();
		while (it.hasNext()) {
			String key = (String) it.next();
			EngineEntityAttributes.put(key, null);
		}
	}
		
	@SuppressWarnings("unchecked")
	public String buildFinalJSON(HashMap<String, String> bwEngineMemAttributesMap,String appName,String agentJMXUrl) 
	{
		JSONObject obj = new JSONObject();
		
		obj.put("timestamp", timeStamp);
		obj.put("ApplicationName", appName);
		obj.put("JMXURL", agentJMXUrl);
		obj.put("MsgType", "BWEngineHealth");
		
		Long heapMemUsed = Long.parseLong(bwEngineMemAttributesMap.get("HeapMemoryUsage-used"));
    	Long heapMemMax = Long.parseLong(bwEngineMemAttributesMap.get("HeapMemoryUsage-max"));
    	Float memUsagePercent = percent(heapMemUsed, heapMemMax);
    	
		String status = (String) bwEngineMemAttributesMap.get("Status");
		if( (status != null || status != "Stopped" ) && memUsagePercent > this.globalProperties.getWarnOnHeapMemUtilizationPercentage())
		{
			//obj.put("Status", "WARN");
			obj.put("Status", status);
			obj.put("Message", "Heap Memory used > " + this.globalProperties.getWarnOnHeapMemUtilizationPercentage() + "%");
		}
		else
		{
			obj.put("Status", status);
			obj.put("Message", bwEngineMemAttributesMap.get("Message"));
		}
		
		obj.put("HeapMemMax", bwEngineMemAttributesMap.get("HeapMemoryUsage-max"));
		obj.put("HeapMemUsed", bwEngineMemAttributesMap.get("HeapMemoryUsage-used"));
		obj.put("Threads", bwEngineMemAttributesMap.get("Threads"));
		obj.put("Uptime", bwEngineMemAttributesMap.get("Uptime"));
		obj.put("Version", bwEngineMemAttributesMap.get("Version"));
		obj.put("ActiveProcessCount", bwEngineMemAttributesMap.get("ActiveProcessCount"));
		obj.put("ProcessCount",bwEngineMemAttributesMap.get("ProcessCount"));
    	
    	return obj.toString();
	}
	
	
	public float percent(Long usedMem, Long totalMem)
	{
		float percentage;
		percentage = (usedMem * 100/ totalMem);
		return percentage;
	}
	
	@SuppressWarnings("unchecked")
	public void buildBWEngineActiveProcessCountJson(Integer activeProcessCount,JSONObject objBWEngineProcess) 
	{
    	objBWEngineProcess.put("ActiveProcessCount", activeProcessCount);
	}
	
	@SuppressWarnings("unchecked")
	public void buildBWEngineProcessCountJson(Integer processCount,JSONObject objBWEngineProcess) 
	{
    	objBWEngineProcess.put("ProcessCount", processCount);
	}
	
	public void publishJMSMessage(List<String> jsons) throws JMSException 
	{
		TopicSession JMSSession = null;
		TopicPublisher JMSPublisher = null;
		Iterator<String> it = jsons.iterator();
		
		try 
		{
			JMSSession = JmsSessionFactory.createJMSSession();
			JMSPublisher = JMSSession.createPublisher(JMSSession.createTopic(MFTopicName));
			
			while (it.hasNext()) 
			{
				TextMessage jmsMessage = JMSSession.createTextMessage();
				while (it.hasNext()) 
				{
					String msg = (String) it.next();
					jmsMessage.setText(msg);
					logger.info("Publishing JMS Json string: " + msg);
					JMSPublisher.send(jmsMessage);
				}
			}
			JMSPublisher.close();
			JMSSession.close();
		} catch (JMSException e) {
			logger.info("Received exception: "
					+ e.getMessage());
		} catch (NamingException e) {
			logger.info("Received exception: "
					+ e.getMessage());
		} catch (InterruptedException e) {
			logger.info("Received exception: "
					+ e.getMessage());
		} finally {
			if (JMSPublisher != null) JMSPublisher.close();
			if (JMSSession != null) JMSSession.close();
		}
	}	

	public Object invokeMbeanOperation(String objectName,String operationName,Object  opParams[], String  opSig[]) 
			throws InstanceNotFoundException, MalformedObjectNameException, MBeanException, ReflectionException, IOException
	{
		Object mbean = JMXcon.getMBeanServerConnection().invoke(new ObjectName(objectName), operationName, opParams, opSig);
		return mbean;
	}

}