package com.apg.eas.main;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import javax.annotation.PostConstruct;
import javax.jms.JMSException;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.MBeanException;
import javax.management.MalformedObjectNameException;
import javax.management.ReflectionException;
import javax.naming.NamingException;

import org.apache.log4j.Logger;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationContext;
import org.springframework.context.annotation.AnnotationConfigApplicationContext;
import org.springframework.context.annotation.ComponentScan;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.io.ClassPathResource;

import com.apg.eas.common.EMSServerDetails;
import com.apg.eas.common.JMXPort;
import com.apg.eas.ems.EMSMonitoringAgent;
import com.apg.eas.ems.EMSMonitoringAgentDetails;
import com.apg.eas.global.properties.GlobalProperties;
import com.apg.eas.jmx.JMXPortMonitoringAgent;
import com.apg.eas.jmx.JMXPortMonitoringAgentDetails;
import com.apg.eas.transport.jms.JmsSessionFactory;
import com.fasterxml.jackson.databind.MappingIterator;
import com.fasterxml.jackson.dataformat.csv.CsvMapper;
import com.fasterxml.jackson.dataformat.csv.CsvSchema;

@Configuration
@ComponentScan(basePackages = "com.apg.eas")
public class Application
{
	//static Logger logger = Utils.getLogger(Application.class);
	private static final Logger logger = Logger.getLogger(Application.class);

	
	@Autowired
	private GlobalProperties globalProperties;
	
	@Autowired
	private List<EMSServerDetails> emsServerDetails;
	
	@Autowired
	private List<JMXPort> jmxPorts;
	
	public static void main(String[] args)
	{
		logger.info("    1. ############################ STARTING THE MONITORING FRAMEWORK...");
		ApplicationContext context = null;
		
		try 
		{
			context = new AnnotationConfigApplicationContext(Application.class);
			Application app = context.getBean(Application.class);
			app.start();
		}
		catch (BeansException e) 
		{
			e.printStackTrace();
		}
	}

	private void start()
	{
		try 
		{
			monitorEMSFromCSV();
			monitorJMXPorts();
			//monitorEMS();
			//monitorBW();
		} 
		catch (InterruptedException e) {
			e.printStackTrace();
		}
		catch (InstanceNotFoundException e) {
			e.printStackTrace();
		} catch (IntrospectionException e) {
			e.printStackTrace();
		} catch (MalformedObjectNameException e) {
			e.printStackTrace();
		} catch (AttributeNotFoundException e) {
			e.printStackTrace();
		} catch (ReflectionException e) {
			e.printStackTrace();
		} catch (MBeanException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (JMSException e) {
			e.printStackTrace();
		}

		while (true) 
		{

		}
	}
	
	public int add(int a,int b) {
		return a+b;
	}


	/*@SuppressWarnings("rawtypes")
	private void monitorBWFromCSV() throws InstanceNotFoundException, IntrospectionException, MalformedObjectNameException, AttributeNotFoundException, ReflectionException, MBeanException, IOException, JMSException, InterruptedException 
	{
		HashMap<String, JMXMonitoringAgentDetails> JMXMonitoringAgents = new HashMap<String, JMXMonitoringAgentDetails>();
		
		for (Iterator iterator = bwConfigs.iterator(); iterator.hasNext();) 
		{
			BWConfig bwConfig = (BWConfig) iterator.next();
			System.out.println("App Name : "+bwConfig.getAppname());
			System.out.println("Instance Name : "+bwConfig.getInstance());
			System.out.println("Host Name : "+bwConfig.getHost());
			System.out.println("Pairing Name : "+bwConfig.getPairing());
			System.out.println("App Name : "+bwConfig.getPort());
			System.out.println("#############################################################################################");
			
			TimerTask monitoringAgent=null;// = new JMXMonitoringAgent(bwConfig.getAppname(), bwConfig.getHost(),"BW_ENGINE",this.globalProperties); 
			Timer monitoringAgentTimer = new Timer(true);
			monitoringAgentTimer.scheduleAtFixedRate(monitoringAgent, 0, globalProperties.getJmxTimeInterval());
			JMXMonitoringAgentDetails monitoringAgentDetails = new JMXMonitoringAgentDetails( monitoringAgent, bwConfig.getHost(), monitoringAgentTimer);
			JMXMonitoringAgents.put(bwConfig.getHost(), monitoringAgentDetails); 
			Thread.sleep(2000);
		}
	}*/
	
	private void monitorJMXPorts() throws InstanceNotFoundException, IntrospectionException, MalformedObjectNameException, AttributeNotFoundException, ReflectionException, MBeanException, IOException, JMSException, InterruptedException 
	{
		HashMap<String, JMXPortMonitoringAgentDetails> JMXMonitoringAgents = new HashMap<String, JMXPortMonitoringAgentDetails>();

		logger.info("    5. ############################ Creating JMX Ports Monitoring Agent");	
		TimerTask monitoringAgent = new JMXPortMonitoringAgent(jmxPorts,globalProperties); 
		Timer monitoringAgentTimer = new Timer(true);
		monitoringAgentTimer.scheduleAtFixedRate(monitoringAgent, 0, globalProperties.getJmxTimeInterval());
		JMXPortMonitoringAgentDetails monitoringAgentDetails = new JMXPortMonitoringAgentDetails( monitoringAgent, "jmxMonitoring", monitoringAgentTimer);
		JMXMonitoringAgents.put("jmxMonitoring", monitoringAgentDetails); 
		Thread.sleep(2000);
	}
	
	@PostConstruct
	private void init() throws NamingException, JMSException 
	{
		logger.info("    2. ############################ GLOBAL CONFIG LOADED WITH BELOW VALUES.");
		logger.info("jNDIProviderContextFactory : " + globalProperties.getjNDIProviderContextFactory());
		logger.info("jNDIProviderURL : " + globalProperties.getjNDIProviderURL());
		logger.info("jNDIUserName : " + globalProperties.getjNDIUserName());
		logger.info("jNDIPassword : " + globalProperties.getjNDIPassword());
		logger.info("jMSUserName : " + globalProperties.getjMSUserName());
		logger.info("jMSPassword : " + globalProperties.getjMSPassword());
		logger.info("topicConnectionFactory : " + globalProperties.getTopicConnectionFactory());
		logger.info("JMS topicName : " + globalProperties.getJmsTopicName());
		logger.info("JMX topicName : " + globalProperties.getJmxTopicName());
		logger.info("fileDirectory : " + globalProperties.getFileDirectory());
		logger.info("fileArchiveDirectory : " + globalProperties.getFileArchiveDirectory());
		logger.info("emsTimeInterval : " + globalProperties.getEmsTimeInterval());
		logger.info("jmxTimeInterval : " + globalProperties.getJmxTimeInterval());
		logger.info("Pub Sub URL : " + globalProperties.getPubSubUrl());
		logger.info("JMX Ports file location : " + globalProperties.getJmxPortsCSVlocation());
		logger.info("######################################  BW ENGINE MBEAN #######################################");
		logger.info("#############################################################################################");
		
		logger.info("    3. ############################ INITIALIZING JMS SESSION FACTORY.");
		JmsSessionFactory.init(
					globalProperties.getjNDIProviderContextFactory(),
					globalProperties.getjNDIProviderURL(),
					globalProperties.getjNDIUserName(),
					globalProperties.getjNDIPassword(),
					globalProperties.getjMSUserName(),
					globalProperties.getjMSPassword(),
					globalProperties.getTopicConnectionFactory());
		logger.info("#############################################################################################");
		
		logger.info("    4. ############################ LOADING JMSPorts.csv AND INITIALIZING JMXPorts.java OBJECTS");
		jmxPorts = loadJMXPorts(JMXPort.class, globalProperties.getJmxPortsCSVlocation());
		emsServerDetails = loadEMSservers(EMSServerDetails.class, globalProperties.getEmsServerDetailslocation());
		logger.info("#############################################################################################");
	}
	
	public <T> List<T> loadBWConfigObjectList(Class<T> type, String fileName) 
	{
	    try
	    {
	        CsvSchema bootstrapSchema = CsvSchema.emptySchema().withHeader();
	        CsvMapper mapper = new CsvMapper();
	        File file = new ClassPathResource(fileName).getFile();
	        MappingIterator<T> readValues = mapper.reader(type).with(bootstrapSchema).readValues(file);
	        return readValues.readAll();
	    }
	    catch (Exception e)
	    {
	    	e.printStackTrace();
	        logger.error("Error occurred while loading object list from file " + fileName, e);
	        return Collections.emptyList();
	    }
	}

	public <T> List<T> loadJMXPorts(Class<T> type, String fileName) 
	{
	    try
	    {
	        CsvSchema bootstrapSchema = CsvSchema.emptySchema().withHeader();
	        CsvMapper mapper = new CsvMapper();
	        File file = new File(fileName);//ClassPathResource(fileName).getFile();
	        MappingIterator<T> readValues = mapper.reader(type).with(bootstrapSchema).readValues(file);
	        return readValues.readAll();
	    }
	    catch (Exception e)
	    {
	    	e.printStackTrace();
	        logger.error("Error occurred while loading object list from file " + fileName, e);
	        return Collections.emptyList();
	    }
	}
	
	public <T> List<T> loadEMSservers(Class<T> type, String fileName) 
	{
	    try
	    {
	        CsvSchema bootstrapSchema = CsvSchema.emptySchema().withHeader();
	        CsvMapper mapper = new CsvMapper();
	        File file = new File(fileName);//ClassPathResource(fileName).getFile();
	        MappingIterator<T> readValues = mapper.reader(type).with(bootstrapSchema).readValues(file);
	        return readValues.readAll();
	    }
	    catch (Exception e)
	    {
	    	e.printStackTrace();
	        logger.error("Error occurred while loading object list from file " + fileName, e);
	        return Collections.emptyList();
	    }
	}
	
	/*@SuppressWarnings("rawtypes")
	private void monitorEMS() throws InterruptedException 
	{
		HashMap<String, EMSMonitoringAgentDetails> EMSMonitoringAgents = new HashMap<String, EMSMonitoringAgentDetails>();
		
		Iterator servers = globalProperties.getEmsServersMap().entrySet().iterator();
		while (servers.hasNext()) 
		{
		    Map.Entry entry = (Map.Entry) servers.next();
		    //String key = (String)entry.getKey();
		    String value = (String)entry.getValue();
		    
		    String emsServerDetails = value;
			String[] tokens = emsServerDetails.split("\\|");

			if (tokens.length != 3)
			{
				logger.error("EmsServerString " + emsServerDetails	+ " is invalid. Correct format is <EMS Server URL>|<admin user>|<admin password>");
				logger.error("Shutting down monitoring framework. Correct the settings and start again.");
				System.exit(0);
			}
			else
			{
				String emsServerURL = tokens[0];
				String userName = tokens[1];
				String password = tokens[2];
				logger.info("Creating EMS Monitoring Agent for : " + emsServerDetails);
				TimerTask monitoringAgent = new EMSMonitoringAgent(emsServerURL, userName, password,globalProperties);
				Timer monitoringAgentTimer = new Timer(true);
				monitoringAgentTimer.scheduleAtFixedRate(monitoringAgent, 0, globalProperties.getEmsTimeInterval());
				EMSMonitoringAgentDetails monitoringAgentDetails = new EMSMonitoringAgentDetails(monitoringAgent, emsServerURL, monitoringAgentTimer);
				EMSMonitoringAgents.put(emsServerURL, monitoringAgentDetails);
				Thread.sleep(2000);
			}
		}
	}*/
	
	@SuppressWarnings("rawtypes")
	private void monitorEMSFromCSV() throws InterruptedException 
	{
		HashMap<String, EMSMonitoringAgentDetails> EMSMonitoringAgents = new HashMap<String, EMSMonitoringAgentDetails>();
		
		for (Iterator iterator = emsServerDetails.iterator(); iterator.hasNext();)
		{
			EMSServerDetails emsServer = (EMSServerDetails) iterator.next();
			
			if (emsServer == null || emsServer.getEmsurl() == "" || emsServer.getUsername() == "" || emsServer.getPwd()=="")
			{
				logger.error("EmsServerString " + emsServerDetails	+ " is invalid. Correct format is <EMS Server URL>|<admin user>|<admin password>");
				logger.error("Shutting down monitoring framework. Correct the settings and start again.");
				System.exit(0);
			}
			else
			{
				String emsServerURL = emsServer.getEmsurl();
				String userName = emsServer.getUsername();
				String password = emsServer.getPwd();
				logger.info("Creating EMS Monitoring Agent for : " + emsServerDetails);
				TimerTask monitoringAgent = new EMSMonitoringAgent(emsServerURL, userName, password,globalProperties);
				Timer monitoringAgentTimer = new Timer(true);
				monitoringAgentTimer.scheduleAtFixedRate(monitoringAgent, 0, globalProperties.getEmsTimeInterval());
				EMSMonitoringAgentDetails monitoringAgentDetails = new EMSMonitoringAgentDetails(monitoringAgent, emsServerURL, monitoringAgentTimer);
				EMSMonitoringAgents.put(emsServerURL, monitoringAgentDetails);
				Thread.sleep(2000);
			}
		}
	}
}