package com.apg.eas.transport.jms;

import java.util.Hashtable;

import javax.jms.JMSException;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicSession;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.slf4j.Logger;

import com.apg.eas.common.Utils;

public class JmsSessionFactory {
	static String InititalContextFactory = null;
	static String ProviderUrl = null;
	static String JNDIUserName = null;
	static String JNDIPassword = null;
	static String JMSUserName = null;
	static String JMSPassword = null;
	static String TopicConnectionFactory = null;
	static TopicConnection topicConnection = null;
	static Logger logger = null;

	public static void init(String initialContextFactory, String providerUrl,String jndiUser, String jndiPass, String jmsUser, String jmsPass, String topicConnectionFactory)
									throws NamingException, JMSException
	{
		InititalContextFactory = initialContextFactory;
		ProviderUrl = providerUrl;
		JNDIUserName = jndiUser;
		JNDIPassword = jndiPass;
		JMSUserName = jmsUser;
		JMSPassword = jmsPass;
		TopicConnectionFactory = topicConnectionFactory;
		logger = Utils.getLogger(JmsSessionFactory.class);
		createConnection();
	}
	
	public static void createConnection() throws NamingException, JMSException
	{
		logger.info("CREATING CONNECTION.");
		/*Init JNDI Context.*/
		Hashtable<String, String> env = new Hashtable<String, String>();
		env.put(Context.INITIAL_CONTEXT_FACTORY, InititalContextFactory);
		env.put(Context.PROVIDER_URL, ProviderUrl);

		if (JNDIUserName != null)
		{
			env.put(Context.SECURITY_PRINCIPAL, JNDIUserName);

			if (JNDIPassword != null)
				env.put(Context.SECURITY_CREDENTIALS, JNDIPassword);
		}

		InitialContext jndiContext = new InitialContext(env);
		logger.info("Initial Context Factory is: " + jndiContext);

		/* Lookup topic connection factory which must exist in the factories config file. */
		TopicConnectionFactory topicFactory = (TopicConnectionFactory) jndiContext.lookup(TopicConnectionFactory);
		logger.info("OK - successfully did lookup TopicConnectionFactory, "	+ topicFactory);

		/*
		 * Let's create a connection to the server and a session to verify the
		 * server is running so we can continue with our lookup operations.
		 */
		topicConnection = topicFactory.createTopicConnection(JMSUserName,JMSPassword);
	}

	public static JmsSessionFactory createTopicPublisher(String topicName) throws JMSException 
	{
		TopicSession topicSession = topicConnection.createTopicSession(false,javax.jms.Session.AUTO_ACKNOWLEDGE);
		javax.jms.Topic topic = topicSession.createTopic(topicName);
		
		JmsSessionFactory publisher = (JmsSessionFactory) topicSession.createPublisher(topic);
		return publisher;
	}

	public static TopicSession createJMSSession() throws JMSException, NamingException, InterruptedException 
	{
		TopicSession session = null;
		for (int i = 0; i <= 3; i++) 
		{
			try
			{
				session = topicConnection.createTopicSession(false,	javax.jms.Session.AUTO_ACKNOWLEDGE);
				break;
			}
			catch (JMSException e)
			{
				logger.info("Received exception. Retrying connection: "+e.getMessage());
//				session.close();
				topicConnection.close();
				Thread.sleep(5000);
				createConnection();
			}
		}
		return session;
	}
}
